#!/bin/sh
#
# Place commands to be executed when WindowMaker is exited here.
#
# WindowMaker will wait until this script finishes, so if you run any
# commands that take long to execute (like a xterm), put a ``&'' in the
# end of the command line.
#
# This file must be executable.
#

